""" persistence class. """

__version__ = 20

## imports

import _thread
import datetime
import importlib
import json
import logging
import os
import queue
import sys
import time
import threading

## defines

homedir = os.path.expanduser("~")
workdir = ""

def locked(func):

    lock = _thread.allocate_lock()

    def lockedfunc(*args, **kwargs):
        lock.acquire()
        res = None
        try:
            res = func(*args, **kwargs)
        finally:
            try:
                lock.release()
            except Exception:
                pass
        return res

    return lockedfunc

## typed helpers

def get_cls(name):
    modname, clsname = name.rsplit(".", 1)
    mod = importlib.import_module(modname)
    return getattr(mod, clsname)

def get_type(o):
    t = type(o)
    if t == type:
        return get_vartype(o)
    else:
        return str(type(o)).split()[-1][1:-2]

def get_clstype(o):
    try:
        return "%s.%s" % (o.__class__.__module__, o.__class__.__name__)
    except AttributeError:
        pass
        
def get_objtype(o):
    try:
        return "%s.%s" % (o.__self__.__module__, o.__self__.__name__)
    except AttributeError:
        pass
         
def get_vartype(o):
    try:
        return "%s.%s" % (o.__module__, o.__name__)
    except AttributeError:
        pass

import ob.utils

## classes

class Object:

    __slots__ = ("__dict__", "_path", "_ready", "_type")

    def __init__(self):
        super().__init__()
        self._path = ""
        self._ready = threading.Event()
        self._type = get_type(self)

    def __eq__(self, o):
        if o:
            return self.__dict__ == o.__dict__

    def __getitem__(self, n, d=None):
        return self.get(n, d)

    def __hash__(self):
        return hash(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __ne__(self, o):
        return self.__dict__ != o.__dict__

    def __iter__(self):
        return iter(self.__dict__)

    def __setitem__(self, k, v):
        self.__dict__.__setitem__(k, v)

    def __str__(self):
        return json.dumps(self.__dict__, default=default, indent=4, sort_keys=True)

    def get(self, k, d=None):
        return getattr(self, k ,d)

    def json(self):
        return json.dumps(self, default=default, indent=4, sort_keys=True)

    def items(self):
        return self.__dict__.items()

    def keys(self):
        return self.__dict__.keys()

    def load(self, path):
        assert path
        assert workdir
        p = os.path.join(workdir, "store", path)
        if not os.path.exists(p):
            assert ob.errors.ENOFILE(path)
        with open(p) as f:
            try:
                d = json.load(f, object_hook=hook)
            except json.decoder.JSONDecodeError:
                raise ob.errors.EJSON(path)
            self.update(d)
        self._path = path
        cache[path] = self 
        return self

    def parse(self):
        pass

    def ready(self):
        self._ready.set()

    @locked
    def save(self, path="", stime=None, timed=False):
        assert workdir
        if self._path and not timed:
            path = self._path
        else:
            t = get_type(self)
            if t not in ob.tables.classes:
                raise ob.errors.ECLASS(t)
            if not stime:
                stime = str(datetime.datetime.now()).replace(" ", os.sep)
            path = os.path.join(t, stime)
            self._path = path
        logging.debug("save %s" % self._path)
        pp = os.path.join(workdir, "store", self._path)
        cdir(pp)
        with open(pp, "w") as f:
            json.dump(self, f, default=default, indent=4, sort_keys=True)
        return path

    def set(self, k, v):
        setattr(self, k, v)

    def update(self, o, keys=[], skip=False):
        if not o:
            return self
        for k, v in o.items():
            if keys and k in keys:
                setattr(self, k, v)
                continue
            if skip and not v:
                continue
            #if skip and self.get(k, None):
            #    continue
            setattr(self, k, v)
        return self

    def wait(self):
        self._ready.wait()

class Cfg(Object):

    def __init__(self, cfg={}):
        super().__init__()
        if cfg:
            self.update(cfg)

    def __getattr__(self, k):
        if k not in dir(self):
            self.set(k, "")
        return self.get(k)

    def last(self):
        from .kernel import k
        l = k.db.last(get_type(self))
        if l:
            self.update(l, skip=True)
        return self

import ob.handler

class Bot(ob.handler.Handler):

    def __init__(self):
        super().__init__()
        self._outputed = False
        self._outqueue = queue.Queue()
        self.cfg = Cfg()
        self.cfg.update({"prompt": True, "verbose": True})
        self.channels = []
        k.fleet.add(self)

    def announce(self, txt):
        for channel in self.channels:
            self.say(channel, txt)

    def output(self):
        self._outputed = True
        while not self._stopped:
            channel, txt, type = self._outqueue.get()
            if txt:
                self.say(channel, txt, type)
            
    def raw(self, txt):
        if not txt:
            return
        sys.stdout.write(txt)
        sys.stdout.flush()
        
    def say(self, channel, txt, type=None):
        if self.cfg.verbose:
            self.raw(str(txt) + "\n")

class Console(Bot):

    def __init__(self):
        super().__init__()
        self._prompted = threading.Event()
        self._prompted.set()

    def dispatch(self, event):
        super().dispatch(event)
        event.wait()
        self._prompted.set()
        return event

    def get_event(self):
        e = ob.event.Event()
        e.direct = True
        e.options = k.cfg.options
        e.origin = "root@shell"
        self._prompted.wait()
        if self.cfg.prompt:
            e.txt = input("> ")
            e.txt = e.txt.rstrip()
        self._prompted.clear()
        return e

class KBot(Bot):

    def dispatch(self, event):
        event.orig = repr(self)
        k.dispatch(event)
        event.wait()
        self._prompted.set()
        return event

import ob.db
import ob.event
import ob.fleet
import ob.user

class Kernel(ob.handler.Handler):

    def __init__(self):
        super().__init__()
        self.cfg = Cfg()
        self.db = ob.db.Cursor()
        self.fleet = ob.fleet.Fleet()
        self.state = Object()
        self.state.starttime = time.time()
        self.state.started = False
        self.users = ob.user.Users()

    def forever(self, name="ob", version=__version__, wd="", usage=""):
        ob.term.termsave()
        try:
            ob.shell.parse_cli(name, version, wd, usage)
            if "--help" in sys.argv:
                return
            if self.cfg.background:
                daemon()
            self.state.started = True
            if not k.cfg.args or k.cfg.modules:
                k.cfg.shell = True
            bot = ob.Console()
            if name and name != "ob":
                k.cfg.modules += ",%s" % name
            bot.initialize(k.cfg.modules)
            if k.cfg.shell:
                bot.start()
                bot.wait()
            else:
                txt = cmd(bot, " ".join(k.cfg.args))
                if txt:
                    print(txt)
        except KeyboardInterrupt:
            print("")
        except Exception as ex:
            logging.error(ob.utils.get_exception())
        ob.term.reset()

## vars

cache = Object()
k = Kernel()

## functions

def cached(path):
    if path not in cache:
        cache[path] = filehook(path)
    return cache[path]

def cdir(path):
    if os.path.exists(path):
        return
    res = ""
    path = os.path.dirname(path)
    for p in path.split(os.sep):
        res += "%s%s" % (p, os.sep)
        padje = os.path.abspath(os.path.normpath(res))
        try:
            os.mkdir(padje)
        except (IsADirectoryError, NotADirectoryError, FileExistsError):
            pass
    return True

def cmd(bot, txt, origin=""):
    e = ob.event.Event()
    e.batch = True
    e.txt = txt
    e.options = k.cfg.options
    e.origin = origin or "root@shell"
    bot.dispatch(e)
    e.wait()
    return "\n".join([str(x).strip() for x in e._result])
       
def default(o):
    t = type(o)
    if isinstance(o, Object):
        return vars(o)
    if isinstance(o, dict):
        return o.items()
    if isinstance(o, list):
        return iter(o)
    if t in [str, True, False, int, float]:
        return o
    return repr(o)

def edit(o, setter={}):
    if not setter:
        return 0
    nr = 0
    for key, value in setter.items():
        if "," in value:
            value = value.split(",")
        t = type(value)
        if value in ["True", "true"]:
            o.set(key, True)
        elif value in ["False", "false"]:
            o.set(key, False)
        elif t == list:
            o.set(key, value)
        elif t == str:
            o.set(key, value)
        else:
            o.set(key, value)
        nr += 1
    return nr

def filehook(fn):
    t = fn.split(os.sep)[0]
    if not t:
        raise ENOFILE(fn)
    o = get_cls(t)()
    return o.load(fn)

def format(o, keys=None, full=False):
    if keys is None:
        keys = vars(self).keys()
    res = []
    txt = ""
    for k in keys:
        v = getattr(o, k, None)
        if k == "text":
            v = v.replace("\\n", "\n")
        if not v:
            continue
        v = str(v)
        if full:
            res.append("%s=%s " % (k, v))
        else:
            res.append(v)
    for v in res:
        txt += "%s " % v.strip()
    return txt.strip()

def hd(*args):
    homedir = os.path.expanduser("~")
    return os.path.abspath(os.path.join(homedir, *args))

def hook(d):
    if "_type" in d:
        t = d["_type"]
        o = get_cls(t)()
    else:
        o = Object()
    o.update(d)
    return o

def kill(thrname):
    for task in threading.enumerate():
        if thrname not in str(task):
            continue
        if "cancel" in dir(task):
            task.cancel()
        if "exit" in dir(task):
            task.exit()
        if "stop" in dir(task):
            task.stop()

def launch(func, *args):
    from ob.tasks import Launcher
    l = Launcher()
    return l.launch(func, *args)

def sliced(o, keys=None):
    oo = Object()
    if not keys:
        keys = o.keys()
    for k in keys:
        try:
            oo[k] = o[k]
        except KeyError:
            pass
    return oo

def search(o, m: None):
    if not m:
        m = {}
    res = False
    for k, v in m.items():
        vv = getattr(o, k, None)
        if vv:
            if v == None:
                res = True
            elif v in str(vv):
                res = True
            else:
                res = False
                break
        else:
            res = False
            break
    return res

def widesearch(o, m: None):
    if not m:
        m = {}
    res = False
    for k, v in m.items():
        vv = getattr(o, k, None)
        if vv and  v in str(vv):
            res = True
            break
    return res

import ob
import ob.clock
import ob.cmds
import ob.command
import ob.db
import ob.entry
import ob.errors
import ob.event
import ob.fleet
import ob.handler
import ob.irc
import ob.loader
import ob.log
import ob.shell
import ob.tables
import ob.tasks
import ob.term
import ob.user
import ob.utils
