""" event handler. """

__version__ = 1

## imports

import inspect
import logging
import os
import pkgutil
import queue
import threading
import ob
import ob.loader
import ob.tables
import ob.tasks

## defines

def __dir__():
    return ("Handler",)

## classes

class Handler(ob.loader.Loader, ob.tasks.Launcher):

    def __init__(self):
        super().__init__()
        self._queue = queue.Queue()
        self._ready = threading.Event()
        self._stopped = False
        self._threaded = True
        self._type = ob.get_type(self)
        self.handlers = {}

    def get_event(self):
        return self._queue.get()

    def get_handler(self, cmd):
        v = self.handlers.get(cmd, None)
        if not v:
            m = ob.tables.modules.get(cmd, None)
            if m:
                self.load_mod(m)
        return self.handlers.get(cmd, None)

    def dispatch(self, event):
        if not event.chk and event.txt:
            event.parse(event.txt)
        if not event.orig:
            event.orig = repr(self)
        try:
            assert event.origin
            event._func = self.get_handler(event.chk)
            if event._func:
                event._func(event)
        except Exception as ex:
            logging.error(ob.utils.get_exception())
        return self.handle_event(event)

    def handle_event(self, event):
        event.ready()
        return event

    def initialize(self, modstr):
        mods = [x for x in modstr.split(",") if x]
        thrs = []
        for mn in mods:
            try:
                self.walk("ob.%s" % mn)
            except ModuleNotFoundError:
                self.walk(mn)
            for modname, m in ob.loader.table.items():
                if mn in modname:
                    if "init" in dir(m):
                        thrs.append(self.launch(m.init))
        for thr in thrs:
            thr.join()
        return thrs

    def loop(self):
        while not self._stopped:
            try:
                event = self.get_event()
            except EOFError:
                break
            if not event:
                continue
            thr = self.launch(self.dispatch, event)
            if self._threaded:
                event._thrs.append(thr)
            else:
                thr.join()

    def load_mod(self, name, mod=None):
        mod = super().load_mod(name, mod)
        self.scan(mod) 
        return mod

    def put(self, event):
        self._queue.put_nowait(event)

    def ready(self):
        self._ready.set()

    def register(self, cmd, handler):
        self.handlers[cmd] = handler

    def scan(self, mod):
        for key, o in inspect.getmembers(mod, inspect.isfunction):
            if "event" in o.__code__.co_varnames:
                self.register(key, o)
                if "modules" not in dir(ob.tables):
                    ob.tables.modules = {}
                ob.tables.modules[key] = o.__module__
        for key, o in inspect.getmembers(mod, inspect.isclass):
            if issubclass(o, ob.Object):
                t = ob.get_type(o)
                if "classes" not in dir(ob.tables):
                    ob.tables.classes = []
                if t not in ob.tables.classes:
                    ob.tables.classes.append(t)
                w = t.split(".")[-1].lower()
                if "names" not in dir(ob.tables):
                    ob.tables.names = {}
                if w not in ob.tables.names:
                    ob.tables.names[w] = t

    def start(self):
        return self.launch(self.loop)

    def stop(self):
        self._stopped = True

    def unload(self, modname):
        mod = self.table.get(modname, None)
        if mod:
            for key, func, name, kind in self.scan(mod):
                if name == modname:
                    try:
                        del self.handlers[key]
                    except KeyError:
                        pass
                    if kind == "class":
                        ob.obj.classes.remove(key)
        super().unload(modname)

    def wait(self):
        self._ready.wait()

    def walk(self, pkgname):
        mod = self.load_mod(pkgname)
        if not mod:
            return
        if "__path__" in dir(mod):
            for n in pkgutil.iter_modules(mod.__path__, mod.__name__+"."):
                self.load_mod(n[1])
        elif "__file__" in dir(mod):
            mn = os.sep.join(mod.__file__.split(os.sep)[:-1])
            for n in pkgutil.walk_packages([mn, ], mod.__name__+"."):
                if n[2]:
                    self.load_mod(n[1])
